/**********************************************************************
 *                                                                    *
 * Graphic Mode Library for Sitronix ST7920 based GLCD                *
 *                                                                    *
 * (c) Axel (XL) Schwenke <axel.schwenke@gmx.net>                     *
 *                                                                    *
 * $Id: st7920.c 245 2014-02-08 22:37:56Z schwenke $
 *                                                                    *
 **********************************************************************/

#include <util/delay.h>

#include "st7920.h"
#include "hal.h"

/* basic delay (half cycle) for the E signal */
#define st7920_delay()   _delay_us(0.6);
/* pulse the E signal L -> H -> L */
#define st7920_E_pulse() st7920_E_H(); st7920_delay(); \
                         st7920_E_L(); st7920_delay();

/* ST7920 commands and flags */
#define DISPLAY_CONTROL  0x08
#define DISPLAY_ON       0x04
#define FUNCTION_SET     0x30
#define EXTENDED_MODE    0x04
#define GRAPHIC_MODE     0x02
#define SET_VMODE        0x02
#define VSCROLL          0x01
#define SET_VSCROLL      0x40
#define SET_GDRAM_ADDR   0x80
#define BUSY_FLAG        0x80


/* module global variables */

/* drawing mode: set, clear, invert */
static st7920_mode mode= set;

/* row-offset(s) of the screen(s) */
static uint8_t draw_offset= 0;
static uint8_t show_offset= 0;


/* inline functions */
static void st7920_check_busy(void)
{
    uint8_t status;
    st7920_port_write(0xFF);
    st7920_port_input();
    st7920_select_command();
    st7920_select_read();

    do {
        st7920_E_H(); st7920_delay();
        status= st7920_port_read();
        st7920_E_L(); st7920_delay();
    } while (status & BUSY_FLAG);

    st7920_port_output();
}

static void st7920_write_command(uint8_t x)
{
    st7920_check_busy();
    st7920_select_command();
    st7920_select_write();
    st7920_port_write(x);
    st7920_E_pulse()
}

static void st7920_set_address(uint8_t row, uint8_t col)
{
    row &= 0x1F;
    col &= 0x0F;

    st7920_check_busy();
    st7920_select_command();
    st7920_select_write();
    st7920_port_write(SET_GDRAM_ADDR + row + draw_offset);
    st7920_E_pulse()

    st7920_check_busy();
    st7920_select_command();
    st7920_select_write();
    st7920_port_write(SET_GDRAM_ADDR + col);
    st7920_E_pulse()

}

static uint8_t st7920_read_data(void)
{
    st7920_check_busy();
    st7920_port_input();
    st7920_select_data();
    st7920_select_read();
    st7920_E_H(); st7920_delay();
    uint8_t x= st7920_port_read();
    st7920_E_L(); st7920_delay();
    st7920_port_output();
    return x;
}

static void st7920_write_data(uint8_t x)
{
    st7920_check_busy();
    st7920_select_data();
    st7920_select_write();
    st7920_port_write(x);
    st7920_E_pulse()
}


/* exported functions */

void st7920_init()
{
    st7920_reset_L();
    _delay_ms(1);
    st7920_reset_H();
    _delay_ms(40);

    st7920_port_output();
    st7920_select_command();
    st7920_select_write();
    st7920_port_write(FUNCTION_SET);
    st7920_E_pulse()
    _delay_us(100);

    st7920_port_write(FUNCTION_SET);
    st7920_E_pulse()

    st7920_write_command(DISPLAY_CONTROL + DISPLAY_ON);
    st7920_write_command(FUNCTION_SET + EXTENDED_MODE);
    st7920_write_command(FUNCTION_SET + EXTENDED_MODE + GRAPHIC_MODE);
    st7920_write_command(SET_VMODE + VSCROLL);
    st7920_write_command(SET_VSCROLL + show_offset);
    st7920_write_command(SET_VMODE);
}


void st7920_set_mode(st7920_mode m)
{
    mode= m;
}


void st7920_set_screen(st7920_screen s)
{
    if (s == front) {
        draw_offset= show_offset;
    } else {
        draw_offset= show_offset ^ 0x20;
    }
}


void st7920_swap(void)
{
    show_offset ^= 0x20;
    draw_offset ^= 0x20;

    st7920_write_command(SET_VMODE + VSCROLL);
    st7920_write_command(SET_VSCROLL + show_offset);
    st7920_write_command(SET_VMODE);
}


void st7920_clear(void)
{
    uint8_t row, col;
    uint8_t val= 0;
    if (mode == clear) {
        val= 0xFF;
    }
    for (row= 0; row < ST7920_HEIGHT; row++) {
        if (row & 0x20) {
            st7920_set_address(row-0x20, 8);
        } else {
            st7920_set_address(row, 0);
        }
        for (col= 0; col < ST7920_WIDTH/8; col++) {
            st7920_write_data(val);
        }
    }
}


void st7920_putpixel(uint8_t x, uint8_t y)
{
    uint16_t mask= 1 << (15 - (x & 0x0F));
    uint8_t col= x/16;
    uint8_t row= y;

    if (row & 0x20) {
        row &= ~0x20;
        col |= 8;
    }

    st7920_set_address(row, col);
    st7920_read_data(); /* dummy read */

    uint16_t data= st7920_read_data();
    data <<= 8;
    data |= st7920_read_data();

    switch (mode)
    {
    case set:
        data |= mask;
        break;
    case clear:
        data &= ~mask;
        break;
    default:
        data ^= mask;
    }

    st7920_set_address(row, col);
    st7920_write_data(data >> 8);
    st7920_write_data(data & 0xFF);
}


//TODO: optimize
void st7920_hline(uint8_t x1, uint8_t y1, uint8_t x2)
{
    for (uint8_t x=x1; x<=x2; x++)
        st7920_putpixel(x, y1);
}


